import logging
import re
from typing import List, Optional, TypedDict

from sqlalchemy import func, select
from sqlalchemy.orm import Session

from ...models import (
    ChartInfo,
    Difficulty,
    DifficultyLocalized,
    Pack,
    Song,
    SongLocalized,
)

logger = logging.getLogger(__name__)


class TArcSongJsonDifficultyItem(TypedDict):
    name_en: str
    name_jp: str
    artist: str
    bpm: str
    bpm_base: float
    set: str
    set_friendly: str
    time: int
    side: int
    world_unlock: bool
    remote_download: bool
    bg: str
    date: int
    version: str
    difficulty: int
    rating: int
    note: int
    chart_designer: str
    jacket_designer: str
    jacket_override: bool
    audio_override: bool


class TArcSongJsonSongItem(TypedDict):
    song_id: str
    difficulties: List[TArcSongJsonDifficultyItem]
    alias: List[str]


class TArcSongJson(TypedDict):
    songs: List[TArcSongJsonSongItem]


class ArcSongJsonBuilder:
    def __init__(self, session: Session):
        self.session = session

    def get_difficulty_item(
        self,
        difficulty: Difficulty,
        song: Song,
        pack: Pack,
        song_localized: Optional[SongLocalized],
    ) -> TArcSongJsonDifficultyItem:
        if "_append_" in pack.id:
            base_pack = self.session.scalar(
                select(Pack).where(Pack.id == re.sub(r"_append_.*$", "", pack.id))
            )
        else:
            base_pack = None

        difficulty_localized = self.session.scalar(
            select(DifficultyLocalized).where(
                (DifficultyLocalized.song_id == difficulty.song_id)
                & (DifficultyLocalized.rating_class == difficulty.rating_class)
            )
        )
        chart_info = self.session.scalar(
            select(ChartInfo).where(
                (ChartInfo.song_id == difficulty.song_id)
                & (ChartInfo.rating_class == difficulty.rating_class)
            )
        )

        if difficulty_localized:
            name_jp = difficulty_localized.title_ja or ""
        elif song_localized:
            name_jp = song_localized.title_ja or ""
        else:
            name_jp = ""

        return {
            "name_en": difficulty.title or song.title,
            "name_jp": name_jp,
            "artist": difficulty.artist or song.artist,
            "bpm": difficulty.bpm or song.bpm or "",
            "bpm_base": difficulty.bpm_base or song.bpm_base or 0.0,
            "set": song.set,
            "set_friendly": f"{base_pack.name} - {pack.name}"
            if base_pack
            else pack.name,
            "time": 0,
            "side": song.side or 0,
            "world_unlock": False,
            "remote_download": False,
            "bg": difficulty.bg or song.bg or "",
            "date": difficulty.date or song.date or 0,
            "version": difficulty.version or song.version or "",
            "difficulty": difficulty.rating * 2 + int(difficulty.rating_plus),
            "rating": chart_info.constant or 0 if chart_info else 0,
            "note": chart_info.notes or 0 if chart_info else 0,
            "chart_designer": difficulty.chart_designer or "",
            "jacket_designer": difficulty.jacket_desginer or "",
            "jacket_override": difficulty.jacket_override,
            "audio_override": difficulty.audio_override,
        }

    def get_song_item(self, song: Song) -> TArcSongJsonSongItem:
        difficulties = self.session.scalars(
            select(Difficulty).where(Difficulty.song_id == song.id)
        )

        pack = self.session.scalar(select(Pack).where(Pack.id == song.set))
        if not pack:
            logger.warning(
                'Cannot find pack "%s", using placeholder instead.', song.set
            )
            pack = Pack(id="unknown", name="Unknown", description="__PLACEHOLDER__")
        song_localized = self.session.scalar(
            select(SongLocalized).where(SongLocalized.id == song.id)
        )

        return {
            "song_id": song.id,
            "difficulties": [
                self.get_difficulty_item(difficulty, song, pack, song_localized)
                for difficulty in difficulties
            ],
            "alias": [],
        }

    def generate_arcsong_json(self) -> TArcSongJson:
        songs = self.session.scalars(select(Song))
        arcsong_songs = []
        for song in songs:
            proceed = self.session.scalar(
                select(func.count(Difficulty.rating_class)).where(
                    Difficulty.song_id == song.id
                )
            )

            if not proceed:
                continue

            arcsong_songs.append(self.get_song_item(song))

        return {"songs": arcsong_songs}



from ...models import Score
from .types import ArcaeaOfflineDEFV2_ScoreItem, ScoreExport


def score(score: Score) -> ScoreExport:
    return {
        "id": score.id,
        "song_id": score.song_id,
        "rating_class": score.rating_class,
        "score": score.score,
        "pure": score.pure,
        "far": score.far,
        "lost": score.lost,
        "date": score.date,
        "max_recall": score.max_recall,
        "modifier": score.modifier,
        "clear_type": score.clear_type,
        "comment": score.comment,
    }


def score_def_v2(score: Score) -> ArcaeaOfflineDEFV2_ScoreItem:
    return {
        "id": score.id,
        "songId": score.song_id,
        "ratingClass": score.rating_class,
        "score": score.score,
        "pure": score.pure,
        "far": score.far,
        "lost": score.lost,
        "date": score.date,
        "maxRecall": score.max_recall,
        "modifier": score.modifier,
        "clearType": score.clear_type,
        "source": None,
        "comment": score.comment,
    }



import logging
import math
from typing import Iterable, List, Optional, Type, Union

from sqlalchemy import Engine, func, inspect, select
from sqlalchemy.orm import DeclarativeBase, InstrumentedAttribute, sessionmaker

from .external.arcsong.arcsong_json import ArcSongJsonBuilder
from .external.exports import ArcaeaOfflineDEFV2_Score, ScoreExport, exporters
from .models.config import ConfigBase, Property
from .models.scores import (
    CalculatedPotential,
    Score,
    ScoreBest,
    ScoreCalculated,
    ScoresBase,
    ScoresViewBase,
)
from .models.songs import (
    Chart,
    ChartInfo,
    Difficulty,
    DifficultyLocalized,
    Pack,
    PackLocalized,
    Song,
    SongLocalized,
    SongsBase,
    SongsViewBase,
)
from .singleton import Singleton

logger = logging.getLogger(__name__)


class Database(metaclass=Singleton):
    def __init__(self, engine: Optional[Engine]):
        try:
            self.__engine
        except AttributeError:
            self.__engine = None

        if engine is None:
            if isinstance(self.engine, Engine):
                return
            raise ValueError("No sqlalchemy.Engine instance specified before.")

        if not isinstance(engine, Engine):
            raise ValueError(
                f"A sqlalchemy.Engine instance expected, not {repr(engine)}"
            )

        if isinstance(self.engine, Engine):
            logger.warning(
                "A sqlalchemy.Engine instance %r has been specified "
                "and will be replaced to %r",
                self.engine,
                engine,
            )
        self.engine = engine

    @property
    def engine(self) -> Engine:
        return self.__engine  # type: ignore

    @engine.setter
    def engine(self, value: Engine):
        if not isinstance(value, Engine):
            raise ValueError("Database.engine only accepts sqlalchemy.Engine")
        self.__engine = value
        self.__sessionmaker = sessionmaker(self.__engine)

    @property
    def sessionmaker(self):
        return self.__sessionmaker

    # region init

    def init(self, checkfirst: bool = True):
        # create tables & views
        if checkfirst:
            # > https://github.com/kvesteri/sqlalchemy-utils/issues/396
            # > view.create_view() causes DuplicateTableError on
            # > Base.metadata.create_all(checkfirst=True)
            # so if `checkfirst` is True, drop these views before creating
            SongsViewBase.metadata.drop_all(self.engine)
            ScoresViewBase.metadata.drop_all(self.engine)

        SongsBase.metadata.create_all(self.engine, checkfirst=checkfirst)
        SongsViewBase.metadata.create_all(self.engine)
        ScoresBase.metadata.create_all(self.engine, checkfirst=checkfirst)
        ScoresViewBase.metadata.create_all(self.engine)
        ConfigBase.metadata.create_all(self.engine, checkfirst=checkfirst)

        # insert version property
        with self.sessionmaker() as session:
            stmt = select(Property.value).where(Property.key == "version")
            result = session.execute(stmt).fetchone()
            if not checkfirst or not result:
                session.add(Property(key="version", value="4"))
                session.commit()

    def check_init(self) -> bool:
        # check table exists
        expect_tables = (
            list(SongsBase.metadata.tables.keys())
            + list(ScoresBase.metadata.tables.keys())
            + list(ConfigBase.metadata.tables.keys())
            + [
                Chart.__tablename__,
                ScoreCalculated.__tablename__,
                ScoreBest.__tablename__,
                CalculatedPotential.__tablename__,
            ]
        )
        return all(inspect(self.engine).has_table(t) for t in expect_tables)

    # endregion

    def version(self) -> Union[int, None]:
        stmt = select(Property).where(Property.key == "version")
        with self.sessionmaker() as session:
            result = session.scalar(stmt)
        return None if result is None else int(result.value)

    # region Pack

    def get_packs(self):
        stmt = select(Pack)
        with self.sessionmaker() as session:
            results = list(session.scalars(stmt))
        return results

    def get_pack(self, pack_id: str):
        stmt = select(Pack).where(Pack.id == pack_id)
        with self.sessionmaker() as session:
            result = session.scalar(stmt)
        return result

    def get_pack_localized(self, pack_id: str):
        stmt = select(PackLocalized).where(PackLocalized.id == pack_id)
        with self.sessionmaker() as session:
            result = session.scalar(stmt)
        return result

    # endregion

    # region Song

    def get_songs(self):
        stmt = select(Song)
        with self.sessionmaker() as session:
            results = list(session.scalars(stmt))
        return results

    def get_songs_by_pack_id(self, pack_id: str):
        stmt = select(Song).where(Song.set == pack_id)
        with self.sessionmaker() as session:
            results = list(session.scalars(stmt))
        return results

    def get_song(self, song_id: str):
        stmt = select(Song).where(Song.id == song_id)
        with self.sessionmaker() as session:
            result = session.scalar(stmt)
        return result

    def get_song_localized(self, song_id: str):
        stmt = select(SongLocalized).where(SongLocalized.id == song_id)
        with self.sessionmaker() as session:
            result = session.scalar(stmt)
        return result

    # endregion

    # region Difficulty

    def get_difficulties(self):
        stmt = select(Difficulty)
        with self.sessionmaker() as session:
            results = list(session.scalars(stmt))
        return results

    def get_difficulties_by_song_id(self, song_id: str):
        stmt = select(Difficulty).where(Difficulty.song_id == song_id)
        with self.sessionmaker() as session:
            results = list(session.scalars(stmt))
        return results

    def get_difficulties_localized_by_song_id(self, song_id: str):
        stmt = select(DifficultyLocalized).where(DifficultyLocalized.song_id == song_id)
        with self.sessionmaker() as session:
            results = list(session.scalars(stmt))
        return results

    def get_difficulty(self, song_id: str, rating_class: int):
        stmt = select(Difficulty).where(
            (Difficulty.song_id == song_id) & (Difficulty.rating_class == rating_class)
        )
        with self.sessionmaker() as session:
            result = session.scalar(stmt)
        return result

    def get_difficulty_localized(self, song_id: str, rating_class: int):
        stmt = select(DifficultyLocalized).where(
            (DifficultyLocalized.song_id == song_id)
            & (DifficultyLocalized.rating_class == rating_class)
        )
        with self.sessionmaker() as session:
            result = session.scalar(stmt)
        return result

    # endregion

    # region ChartInfo

    def get_chart_infos(self):
        stmt = select(ChartInfo)
        with self.sessionmaker() as session:
            results = list(session.scalars(stmt))
        return results

    def get_chart_infos_by_song_id(self, song_id: str):
        stmt = select(ChartInfo).where(ChartInfo.song_id == song_id)
        with self.sessionmaker() as session:
            results = list(session.scalars(stmt))
        return results

    def get_chart_info(self, song_id: str, rating_class: int):
        stmt = select(ChartInfo).where(
            (ChartInfo.song_id == song_id) & (ChartInfo.rating_class == rating_class)
        )
        with self.sessionmaker() as session:
            result = session.scalar(stmt)
        return result

    # endregion

    # region Chart

    def get_charts_by_pack_id(self, pack_id: str):
        stmt = select(Chart).where(Chart.set == pack_id)
        with self.sessionmaker() as session:
            results = list(session.scalars(stmt))
        return results

    def get_charts_by_song_id(self, song_id: str):
        stmt = select(Chart).where(Chart.song_id == song_id)
        with self.sessionmaker() as session:
            results = list(session.scalars(stmt))
        return results

    def get_charts_by_constant(self, constant: int):
        stmt = select(Chart).where(Chart.constant == constant)
        with self.sessionmaker() as session:
            results = list(session.scalars(stmt))
        return results

    def get_chart(self, song_id: str, rating_class: int):
        stmt = select(Chart).where(
            (Chart.song_id == song_id) & (Chart.rating_class == rating_class)
        )
        with self.sessionmaker() as session:
            result = session.scalar(stmt)
        return result

    # endregion

    # region Score

    def get_scores(self):
        stmt = select(Score)
        with self.sessionmaker() as session:
            results = list(session.scalars(stmt))
        return results

    def get_score(self, score_id: int):
        stmt = select(Score).where(Score.id == score_id)
        with self.sessionmaker() as session:
            result = session.scalar(stmt)
        return result

    def get_score_best(self, song_id: str, rating_class: int):
        stmt = select(ScoreBest).where(
            (ScoreBest.song_id == song_id) & (ScoreBest.rating_class == rating_class)
        )
        with self.sessionmaker() as session:
            result = session.scalar(stmt)
        return result

    def insert_score(self, score: Score):
        with self.sessionmaker() as session:
            session.add(score)
            session.commit()

    def insert_scores(self, scores: Iterable[Score]):
        with self.sessionmaker() as session:
            session.add_all(scores)
            session.commit()

    def update_score(self, score: Score):
        if score.id is None:
            raise ValueError(
                "Cannot determine which score to update, please specify `score.id`"
            )
        with self.sessionmaker() as session:
            session.merge(score)
            session.commit()

    def delete_score(self, score: Score):
        with self.sessionmaker() as session:
            session.delete(score)
            session.commit()

    def recommend_charts(self, play_result: float, bounds: float = 0.1):
        base_constant = math.ceil(play_result * 10)

        results = []
        results_id = []
        with self.sessionmaker() as session:
            for constant in range(base_constant - 20, base_constant + 1):
                # from Pure Memory(EX+) to AA
                score_modifier = (play_result * 10 - constant) / 10
                if score_modifier >= 2.0:
                    min_score = 10000000
                elif score_modifier >= 1.0:
                    min_score = 200000 * (score_modifier - 1) + 9800000
                else:
                    min_score = 300000 * score_modifier + 9500000
                min_score = int(min_score)

                charts = self.get_charts_by_constant(constant)
                for chart in charts:
                    score_best_stmt = select(ScoreBest).where(
                        (ScoreBest.song_id == chart.song_id)
                        & (ScoreBest.rating_class == chart.rating_class)
                        & (ScoreBest.score >= min_score)
                        & (play_result - bounds < ScoreBest.potential)
                        & (ScoreBest.potential < play_result + bounds)
                    )
                    if session.scalar(score_best_stmt):
                        chart_id = f"{chart.song_id},{chart.rating_class}"
                        if chart_id not in results_id:
                            results.append(chart)
                            results_id.append(chart_id)

        return results

    # endregion

    def get_b30(self):
        stmt = select(CalculatedPotential.b30).select_from(CalculatedPotential)
        with self.sessionmaker() as session:
            result = session.scalar(stmt)
        return result

    # region COUNT

    def __count_table(self, base: Type[DeclarativeBase]):
        stmt = select(func.count()).select_from(base)
        with self.sessionmaker() as session:
            result = session.scalar(stmt)
        return result or 0

    def __count_column(self, column: InstrumentedAttribute):
        stmt = select(func.count(column))
        with self.sessionmaker() as session:
            result = session.scalar(stmt)
        return result or 0

    def count_packs(self):
        return self.__count_column(Pack.id)

    def count_songs(self):
        return self.__count_column(Song.id)

    def count_difficulties(self):
        return self.__count_table(Difficulty)

    def count_chart_infos(self):
        return self.__count_table(ChartInfo)

    def count_complete_chart_infos(self):
        stmt = (
            select(func.count())
            .select_from(ChartInfo)
            .where((ChartInfo.constant != None) & (ChartInfo.notes != None))
        )
        with self.sessionmaker() as session:
            result = session.scalar(stmt)
        return result or 0

    def count_charts(self):
        return self.__count_table(Chart)

    def count_scores(self):
        return self.__count_column(Score.id)

    def count_scores_calculated(self):
        return self.__count_table(ScoreCalculated)

    def count_scores_best(self):
        return self.__count_table(ScoreBest)

    # endregion

    # region export

    def export_scores(self) -> List[ScoreExport]:
        scores = self.get_scores()
        return [exporters.score(score) for score in scores]

    def export_scores_def_v2(self) -> ArcaeaOfflineDEFV2_Score:
        scores = self.get_scores()
        return {
            "$schema": "https://arcaeaoffline.sevive.xyz/schemas/def/v2/score.schema.json",
            "type": "score",
            "version": 2,
            "scores": [exporters.score_def_v2(score) for score in scores],
        }

    def generate_arcsong(self):
        with self.sessionmaker() as session:
            arcsong = ArcSongJsonBuilder(session).generate_arcsong_json()
        return arcsong

    # endregion



import sqlite3
from typing import List

from sqlalchemy.orm import Session

from ...models.songs import ChartInfo


class ArcsongDbParser:
    def __init__(self, filepath):
        self.filepath = filepath

    def parse(self) -> List[ChartInfo]:
        results = []
        with sqlite3.connect(self.filepath) as conn:
            cursor = conn.cursor()
            arcsong_db_results = cursor.execute(
                "SELECT song_id, rating_class, rating, note FROM charts"
            )
            for result in arcsong_db_results:
                chart = ChartInfo(
                    song_id=result[0],
                    rating_class=result[1],
                    constant=result[2],
                    notes=result[3] or None,
                )
                results.append(chart)

        return results

    def write_database(self, session: Session):
        results = self.parse()
        for result in results:
            session.merge(result)



import json
from typing import List, Union

from ...models.songs import Pack, PackLocalized
from .common import ArcaeaParser, is_localized, set_model_localized_attrs


class PacklistParser(ArcaeaParser):
    def parse(self) -> List[Union[Pack, PackLocalized]]:
        packlist_json_root = json.loads(self.read_file_text())

        packlist_json = packlist_json_root["packs"]
        results: List[Union[Pack, PackLocalized]] = [
            Pack(id="single", name="Memory Archive")
        ]
        for item in packlist_json:
            pack = Pack()
            pack.id = item["id"]
            pack.name = item["name_localized"]["en"]
            pack.description = item["description_localized"]["en"] or None
            results.append(pack)

            if is_localized(item, "name") or is_localized(item, "description"):
                pack_localized = PackLocalized(id=pack.id)
                set_model_localized_attrs(pack_localized, item, "name")
                set_model_localized_attrs(pack_localized, item, "description")
                results.append(pack_localized)

        return results



from .b30 import calculate_b30, get_b30_calculated_list
from .score import (
    calculate_constants_from_play_rating,
    calculate_play_rating,
    calculate_score_modifier,
    calculate_score_range,
    calculate_shiny_pure,
)



from typing import List, Literal, Optional, TypedDict


class ScoreExport(TypedDict):
    id: int
    song_id: str
    rating_class: int
    score: int
    pure: Optional[int]
    far: Optional[int]
    lost: Optional[int]
    date: Optional[int]
    max_recall: Optional[int]
    modifier: Optional[int]
    clear_type: Optional[int]
    comment: Optional[str]


class ArcaeaOfflineDEFV2_ScoreItem(TypedDict, total=False):
    id: Optional[int]
    songId: str
    ratingClass: int
    score: int
    pure: Optional[int]
    far: Optional[int]
    lost: Optional[int]
    date: Optional[int]
    maxRecall: Optional[int]
    modifier: Optional[int]
    clearType: Optional[int]
    source: Optional[str]
    comment: Optional[str]


ArcaeaOfflineDEFV2_Score = TypedDict(
    "ArcaeaOfflineDEFV2_Score",
    {
        "$schema": Literal[
            "https://arcaeaoffline.sevive.xyz/schemas/def/v2/score.schema.json"
        ],
        "type": Literal["score"],
        "version": Literal[2],
        "scores": List[ArcaeaOfflineDEFV2_ScoreItem],
    },
)



import contextlib
import sqlite3
from typing import List

from sqlalchemy.orm import Session

from ...models.songs import ChartInfo


class ChartInfoDbParser:
    def __init__(self, filepath):
        self.filepath = filepath

    def parse(self) -> List[ChartInfo]:
        results = []
        with sqlite3.connect(self.filepath) as conn:
            with contextlib.closing(conn.cursor()) as cursor:
                db_results = cursor.execute(
                    "SELECT song_id, rating_class, constant, notes FROM charts_info"
                ).fetchall()
                for result in db_results:
                    chart = ChartInfo(
                        song_id=result[0],
                        rating_class=result[1],
                        constant=result[2],
                        notes=result[3] or None,
                    )
                    results.append(chart)

        return results

    def write_database(self, session: Session):
        results = self.parse()
        for result in results:
            session.merge(result)



from decimal import Decimal
from typing import Dict, List

from ..models.scores import ScoreCalculated


def get_b30_calculated_list(
    calculated_list: List[ScoreCalculated],
) -> List[ScoreCalculated]:
    best_scores: Dict[str, ScoreCalculated] = {}
    for calculated in calculated_list:
        key = f"{calculated.song_id}_{calculated.rating_class}"
        stored = best_scores.get(key)
        if stored and stored.score < calculated.score or not stored:
            best_scores[key] = calculated
    ret_list = list(best_scores.values())
    ret_list = sorted(ret_list, key=lambda c: c.potential, reverse=True)[:30]
    return ret_list


def calculate_b30(calculated_list: List[ScoreCalculated]) -> Decimal:
    ptt_list = [Decimal(c.potential) for c in get_b30_calculated_list(calculated_list)]
    sum_ptt_list = sum(ptt_list)
    return (sum_ptt_list / len(ptt_list)) if sum_ptt_list else Decimal("0.0")



from typing import List, Union

from sqlalchemy import select
from sqlalchemy.orm import Session
from whoosh.analysis import NgramFilter, StandardAnalyzer
from whoosh.fields import ID, KEYWORD, TEXT, Schema
from whoosh.filedb.filestore import RamStorage
from whoosh.qparser import FuzzyTermPlugin, MultifieldParser, OrGroup

from .models.songs import Song, SongLocalized
from .utils.search_title import recover_search_title


class Searcher:
    def __init__(self):
        self.text_analyzer = StandardAnalyzer() | NgramFilter(minsize=2, maxsize=5)
        self.song_schema = Schema(
            song_id=ID(stored=True, unique=True),
            title=TEXT(analyzer=self.text_analyzer, spelling=True),
            artist=TEXT(analyzer=self.text_analyzer, spelling=True),
            source=TEXT(analyzer=self.text_analyzer, spelling=True),
            keywords=KEYWORD(lowercase=True, stored=True, scorable=True),
        )
        self.storage = RamStorage()
        self.index = self.storage.create_index(self.song_schema)

        self.default_query_parser = MultifieldParser(
            ["song_id", "title", "artist", "source", "keywords"],
            self.song_schema,
            group=OrGroup,
        )
        self.default_query_parser.add_plugin(FuzzyTermPlugin())

    def import_songs(self, session: Session):
        writer = self.index.writer()
        songs = list(session.scalars(select(Song)))
        song_localize_stmt = select(SongLocalized)
        for song in songs:
            stmt = song_localize_stmt.where(SongLocalized.id == song.id)
            sl = session.scalar(stmt)
            song_id = song.id
            possible_titles: List[Union[str, None]] = [song.title]
            possible_artists: List[Union[str, None]] = [song.artist]
            possible_sources: List[Union[str, None]] = [song.source]
            if sl:
                possible_titles.extend(
                    [sl.title_ja, sl.title_ko, sl.title_zh_hans, sl.title_zh_hant]
                )
                possible_titles.extend(
                    recover_search_title(sl.search_title_ja)
                    + recover_search_title(sl.search_title_ko)
                    + recover_search_title(sl.search_title_zh_hans)
                    + recover_search_title(sl.search_title_zh_hant)
                )
                possible_artists.extend(
                    recover_search_title(sl.search_artist_ja)
                    + recover_search_title(sl.search_artist_ko)
                    + recover_search_title(sl.search_artist_zh_hans)
                    + recover_search_title(sl.search_artist_zh_hant)
                )
                possible_sources.extend(
                    [
                        sl.source_ja,
                        sl.source_ko,
                        sl.source_zh_hans,
                        sl.source_zh_hant,
                    ]
                )

            # remove empty items in list
            titles = [t for t in possible_titles if t != "" and t is not None]
            artists = [t for t in possible_artists if t != "" and t is not None]
            sources = [t for t in possible_sources if t != "" and t is not None]

            writer.update_document(
                song_id=song_id,
                title=" ".join(titles),
                artist=" ".join(artists),
                source=" ".join(sources),
                keywords=" ".join([song_id] + titles + artists + sources),
            )

        writer.commit()

    def did_you_mean(self, string: str):
        results = set()

        with self.index.searcher() as searcher:
            corrector_keywords = searcher.corrector("keywords")  # type: ignore
            corrector_song_id = searcher.corrector("song_id")  # type: ignore
            corrector_title = searcher.corrector("title")  # type: ignore
            corrector_artist = searcher.corrector("artist")  # type: ignore
            corrector_source = searcher.corrector("source")  # type: ignore

            results.update(corrector_keywords.suggest(string))
            results.update(corrector_song_id.suggest(string))
            results.update(corrector_title.suggest(string))
            results.update(corrector_artist.suggest(string))
            results.update(corrector_source.suggest(string))

        if string in results:
            results.remove(string)

        return list(results)

    def search(self, string: str, *, limit: int = 10):
        query_string = f"{string}"
        query = self.default_query_parser.parse(query_string)
        with self.index.searcher() as searcher:
            results = searcher.search(query, limit=limit)
            return [result.get("song_id") for result in results]



# pylint: disable=too-few-public-methods, duplicate-code

from typing import Optional

from sqlalchemy import TEXT, ForeignKey, func, select
from sqlalchemy.orm import DeclarativeBase, Mapped, mapped_column
from sqlalchemy_utils import create_view

from .common import ReprHelper

__all__ = [
    "SongsBase",
    "Pack",
    "PackLocalized",
    "Song",
    "SongLocalized",
    "Difficulty",
    "DifficultyLocalized",
    "ChartInfo",
    "SongsViewBase",
    "Chart",
]


class SongsBase(DeclarativeBase, ReprHelper):
    pass


class Pack(SongsBase):
    __tablename__ = "packs"

    id: Mapped[str] = mapped_column(TEXT(), primary_key=True)
    name: Mapped[str] = mapped_column(TEXT())
    description: Mapped[Optional[str]] = mapped_column(TEXT())


class PackLocalized(SongsBase):
    __tablename__ = "packs_localized"

    id: Mapped[str] = mapped_column(ForeignKey("packs.id"), primary_key=True)
    name_ja: Mapped[Optional[str]] = mapped_column(TEXT())
    name_ko: Mapped[Optional[str]] = mapped_column(TEXT())
    name_zh_hans: Mapped[Optional[str]] = mapped_column(TEXT())
    name_zh_hant: Mapped[Optional[str]] = mapped_column(TEXT())
    description_ja: Mapped[Optional[str]] = mapped_column(TEXT())
    description_ko: Mapped[Optional[str]] = mapped_column(TEXT())
    description_zh_hans: Mapped[Optional[str]] = mapped_column(TEXT())
    description_zh_hant: Mapped[Optional[str]] = mapped_column(TEXT())


class Song(SongsBase):
    __tablename__ = "songs"

    idx: Mapped[int]
    id: Mapped[str] = mapped_column(TEXT(), primary_key=True)
    title: Mapped[str] = mapped_column(TEXT())
    artist: Mapped[str] = mapped_column(TEXT())
    set: Mapped[str] = mapped_column(TEXT())
    bpm: Mapped[Optional[str]] = mapped_column(TEXT())
    bpm_base: Mapped[Optional[float]]
    audio_preview: Mapped[Optional[int]]
    audio_preview_end: Mapped[Optional[int]]
    side: Mapped[Optional[int]]
    version: Mapped[Optional[str]] = mapped_column(TEXT())
    date: Mapped[Optional[int]]
    bg: Mapped[Optional[str]] = mapped_column(TEXT())
    bg_inverse: Mapped[Optional[str]] = mapped_column(TEXT())
    bg_day: Mapped[Optional[str]] = mapped_column(TEXT())
    bg_night: Mapped[Optional[str]] = mapped_column(TEXT())
    source: Mapped[Optional[str]] = mapped_column(TEXT())
    source_copyright: Mapped[Optional[str]] = mapped_column(TEXT())


class SongLocalized(SongsBase):
    __tablename__ = "songs_localized"

    id: Mapped[str] = mapped_column(ForeignKey("songs.id"), primary_key=True)
    title_ja: Mapped[Optional[str]] = mapped_column(TEXT())
    title_ko: Mapped[Optional[str]] = mapped_column(TEXT())
    title_zh_hans: Mapped[Optional[str]] = mapped_column(TEXT())
    title_zh_hant: Mapped[Optional[str]] = mapped_column(TEXT())
    search_title_ja: Mapped[Optional[str]] = mapped_column(TEXT(), comment="JSON array")
    search_title_ko: Mapped[Optional[str]] = mapped_column(TEXT(), comment="JSON array")
    search_title_zh_hans: Mapped[Optional[str]] = mapped_column(
        TEXT(), comment="JSON array"
    )
    search_title_zh_hant: Mapped[Optional[str]] = mapped_column(
        TEXT(), comment="JSON array"
    )
    search_artist_ja: Mapped[Optional[str]] = mapped_column(
        TEXT(), comment="JSON array"
    )
    search_artist_ko: Mapped[Optional[str]] = mapped_column(
        TEXT(), comment="JSON array"
    )
    search_artist_zh_hans: Mapped[Optional[str]] = mapped_column(
        TEXT(), comment="JSON array"
    )
    search_artist_zh_hant: Mapped[Optional[str]] = mapped_column(
        TEXT(), comment="JSON array"
    )
    source_ja: Mapped[Optional[str]] = mapped_column(TEXT())
    source_ko: Mapped[Optional[str]] = mapped_column(TEXT())
    source_zh_hans: Mapped[Optional[str]] = mapped_column(TEXT())
    source_zh_hant: Mapped[Optional[str]] = mapped_column(TEXT())


class Difficulty(SongsBase):
    __tablename__ = "difficulties"

    song_id: Mapped[str] = mapped_column(TEXT(), primary_key=True)
    rating_class: Mapped[int] = mapped_column(primary_key=True)
    rating: Mapped[int]
    rating_plus: Mapped[bool]
    chart_designer: Mapped[Optional[str]] = mapped_column(TEXT())
    jacket_desginer: Mapped[Optional[str]] = mapped_column(TEXT())
    audio_override: Mapped[bool]
    jacket_override: Mapped[bool]
    jacket_night: Mapped[Optional[str]] = mapped_column(TEXT())
    title: Mapped[Optional[str]] = mapped_column(TEXT())
    artist: Mapped[Optional[str]] = mapped_column(TEXT())
    bg: Mapped[Optional[str]] = mapped_column(TEXT())
    bg_inverse: Mapped[Optional[str]] = mapped_column(TEXT())
    bpm: Mapped[Optional[str]] = mapped_column(TEXT())
    bpm_base: Mapped[Optional[float]]
    version: Mapped[Optional[str]] = mapped_column(TEXT())
    date: Mapped[Optional[int]]


class DifficultyLocalized(SongsBase):
    __tablename__ = "difficulties_localized"

    song_id: Mapped[str] = mapped_column(
        ForeignKey("difficulties.song_id"), primary_key=True
    )
    rating_class: Mapped[str] = mapped_column(
        ForeignKey("difficulties.rating_class"), primary_key=True
    )
    title_ja: Mapped[Optional[str]] = mapped_column(TEXT())
    title_ko: Mapped[Optional[str]] = mapped_column(TEXT())
    title_zh_hans: Mapped[Optional[str]] = mapped_column(TEXT())
    title_zh_hant: Mapped[Optional[str]] = mapped_column(TEXT())
    artist_ja: Mapped[Optional[str]] = mapped_column(TEXT())
    artist_ko: Mapped[Optional[str]] = mapped_column(TEXT())
    artist_zh_hans: Mapped[Optional[str]] = mapped_column(TEXT())
    artist_zh_hant: Mapped[Optional[str]] = mapped_column(TEXT())


class ChartInfo(SongsBase):
    __tablename__ = "charts_info"

    song_id: Mapped[str] = mapped_column(
        ForeignKey("difficulties.song_id"), primary_key=True
    )
    rating_class: Mapped[str] = mapped_column(
        ForeignKey("difficulties.rating_class"), primary_key=True
    )
    constant: Mapped[int] = mapped_column(
        comment="real_constant * 10. For example, Crimson Throne [FTR] is 10.4, then store 104."
    )
    notes: Mapped[Optional[int]]


class SongsViewBase(DeclarativeBase, ReprHelper):
    pass


class Chart(SongsViewBase):
    __tablename__ = "charts"

    song_idx: Mapped[int]
    song_id: Mapped[str]
    rating_class: Mapped[int]
    rating: Mapped[int]
    rating_plus: Mapped[bool]
    title: Mapped[str]
    artist: Mapped[str]
    set: Mapped[str]
    bpm: Mapped[Optional[str]]
    bpm_base: Mapped[Optional[float]]
    audio_preview: Mapped[Optional[int]]
    audio_preview_end: Mapped[Optional[int]]
    side: Mapped[Optional[int]]
    version: Mapped[Optional[str]]
    date: Mapped[Optional[int]]
    bg: Mapped[Optional[str]]
    bg_inverse: Mapped[Optional[str]]
    bg_day: Mapped[Optional[str]]
    bg_night: Mapped[Optional[str]]
    source: Mapped[Optional[str]]
    source_copyright: Mapped[Optional[str]]
    chart_designer: Mapped[Optional[str]]
    jacket_desginer: Mapped[Optional[str]]
    audio_override: Mapped[bool]
    jacket_override: Mapped[bool]
    jacket_night: Mapped[Optional[str]]
    constant: Mapped[int]
    notes: Mapped[Optional[int]]

    __table__ = create_view(
        name=__tablename__,
        selectable=select(
            Song.idx.label("song_idx"),
            Difficulty.song_id,
            Difficulty.rating_class,
            Difficulty.rating,
            Difficulty.rating_plus,
            func.coalesce(Difficulty.title, Song.title).label("title"),
            func.coalesce(Difficulty.artist, Song.artist).label("artist"),
            Song.set,
            func.coalesce(Difficulty.bpm, Song.bpm).label("bpm"),
            func.coalesce(Difficulty.bpm_base, Song.bpm_base).label("bpm_base"),
            Song.audio_preview,
            Song.audio_preview_end,
            Song.side,
            func.coalesce(Difficulty.version, Song.version).label("version"),
            func.coalesce(Difficulty.date, Song.date).label("date"),
            func.coalesce(Difficulty.bg, Song.bg).label("bg"),
            func.coalesce(Difficulty.bg_inverse, Song.bg_inverse).label("bg_inverse"),
            Song.bg_day,
            Song.bg_night,
            Song.source,
            Song.source_copyright,
            Difficulty.chart_designer,
            Difficulty.jacket_desginer,
            Difficulty.audio_override,
            Difficulty.jacket_override,
            Difficulty.jacket_night,
            ChartInfo.constant,
            ChartInfo.notes,
        )
        .select_from(Difficulty)
        .join(
            ChartInfo,
            (Difficulty.song_id == ChartInfo.song_id)
            & (Difficulty.rating_class == ChartInfo.rating_class),
        )
        .join(Song, Difficulty.song_id == Song.id),
        metadata=SongsViewBase.metadata,
        cascade_on_drop=False,
    )



from sqlalchemy.orm import Session

from ...models import Chart, ScoreBest
from ...utils.rating import rating_class_to_text


class SmartRteB30CsvConverter:
    CSV_ROWS = [
        "songname",
        "songId",
        "Difficulty",
        "score",
        "Perfect",
        "criticalPerfect",
        "Far",
        "Lost",
        "Constant",
        "singlePTT",
    ]

    def __init__(
        self,
        session: Session,
    ):
        self.session = session

    def rows(self) -> list:
        csv_rows = [self.CSV_ROWS.copy()]

        with self.session as session:
            results = (
                session.query(
                    Chart.title,
                    ScoreBest.song_id,
                    ScoreBest.rating_class,
                    ScoreBest.score,
                    ScoreBest.pure,
                    ScoreBest.shiny_pure,
                    ScoreBest.far,
                    ScoreBest.lost,
                    Chart.constant,
                    ScoreBest.potential,
                )
                .join(
                    Chart,
                    (Chart.song_id == ScoreBest.song_id)
                    & (Chart.rating_class == ScoreBest.rating_class),
                )
                .all()
            )

            for result in results:
                # replace the comma in song title because the target project
                # cannot handle quoted string
                result = list(result)
                result[0] = result[0].replace(",", "")
                result[2] = rating_class_to_text(result[2])
                # divide constant to float
                result[-2] = result[-2] / 10
                # round potential
                result[-1] = round(result[-1], 5)
                csv_rows.append(result)

        return csv_rows



from typing import Generic, TypeVar

T = TypeVar("T")


class Singleton(type, Generic[T]):
    _instance = None

    def __call__(cls, *args, **kwargs) -> T:
        if cls._instance is None:
            cls._instance = super().__call__(*args, **kwargs)
        return cls._instance



from datetime import datetime
from enum import IntEnum


class KanaeDayNight(IntEnum):
    Day = 0
    Night = 1


def kanae_day_night(timestamp: int) -> KanaeDayNight:
    """
    :param timestamp: POSIX timestamp, which is passed to `datetime.fromtimestamp(timestamp)`.
    """
    dt = datetime.fromtimestamp(timestamp)
    return KanaeDayNight.Day if 6 <= dt.hour <= 19 else KanaeDayNight.Night



import json
import logging
from datetime import datetime
from typing import Dict, List, Literal, Optional, TypedDict

from ...models import Score
from .common import ArcaeaParser, fix_timestamp

logger = logging.getLogger(__name__)


class TWebApiRatingMeScoreItem(TypedDict):
    song_id: str
    difficulty: int
    modifier: int
    rating: float
    score: int
    perfect_count: int
    near_count: int
    miss_count: int
    clear_type: int
    title: Dict[Literal["ja", "en"], str]
    artist: str
    time_played: int
    bg: str


class TWebApiRatingMeValue(TypedDict):
    best_rated_scores: List[TWebApiRatingMeScoreItem]
    recent_rated_scores: List[TWebApiRatingMeScoreItem]


class TWebApiRatingMeResult(TypedDict):
    success: bool
    error_code: Optional[int]
    value: Optional[TWebApiRatingMeValue]


class ArcaeaOnlineParser(ArcaeaParser):
    def parse(self) -> List[Score]:
        api_result_root: TWebApiRatingMeResult = json.loads(self.read_file_text())

        api_result_value = api_result_root.get("value")
        if not api_result_value:
            error_code = api_result_root.get("error_code")
            raise ValueError(f"Cannot parse API result, error code {error_code}")

        best30_score_items = api_result_value.get("best_rated_scores", [])
        recent_score_items = api_result_value.get("recent_rated_scores", [])
        score_items = best30_score_items + recent_score_items

        date_text = datetime.now().strftime("%Y-%m-%d %H:%M:%S")

        results: List[Score] = []
        for score_item in score_items:
            score = Score()
            score.song_id = score_item["song_id"]
            score.rating_class = score_item["difficulty"]
            score.score = score_item["score"]
            score.pure = score_item["perfect_count"]
            score.far = score_item["near_count"]
            score.lost = score_item["miss_count"]
            score.date = fix_timestamp(int(score_item["time_played"] / 1000))
            score.modifier = score_item["modifier"]
            score.clear_type = score_item["clear_type"]

            if score.lost == 0:
                score.max_recall = score.pure + score.far

            score.comment = f"Parsed from web API at {date_text}"
            results.append(score)
        return results



from typing import Any, Sequence

SCORE_GRADE_FLOOR = [9900000, 9800000, 9500000, 9200000, 8900000, 8600000, 0]
SCORE_GRADE_TEXTS = ["EX+", "EX", "AA", "A", "B", "C", "D"]
MODIFIER_TEXTS = ["NORMAL", "EASY", "HARD"]
CLEAR_TYPE_TEXTS = [
    "TRACK LOST",
    "NORMAL CLEAR",
    "FULL RECALL",
    "PURE MEMORY",
    "EASY CLEAR",
    "HARD CLEAR",
]


def zip_score_grade(score: int, __seq: Sequence, default: Any = "__PRESERVE__"):
    """
    zip_score_grade is a simple wrapper that equals to:
    ```py
    for score_floor, val in zip(SCORE_GRADE_FLOOR, __seq):
        if score >= score_floor:
            return val
    return seq[-1] if default == "__PRESERVE__" else default
    ```
    Could be useful in specific cases.
    """
    return next(
        (
            val
            for score_floor, val in zip(SCORE_GRADE_FLOOR, __seq)
            if score >= score_floor
        ),
        __seq[-1] if default == "__PRESERVE__" else default,
    )


def score_to_grade_text(score: int) -> str:
    return zip_score_grade(score, SCORE_GRADE_TEXTS)


def modifier_to_text(modifier: int) -> str:
    return MODIFIER_TEXTS[modifier]


def clear_type_to_text(clear_type: int) -> str:
    return CLEAR_TYPE_TEXTS[clear_type]



import json
from typing import List, Union

from ...models.songs import Difficulty, DifficultyLocalized, Song, SongLocalized
from .common import ArcaeaParser, is_localized, set_model_localized_attrs, to_db_value


class SonglistParser(ArcaeaParser):
    def parse(
        self,
    ) -> List[Union[Song, SongLocalized, Difficulty, DifficultyLocalized]]:
        songlist_json_root = json.loads(self.read_file_text())

        songlist_json = songlist_json_root["songs"]
        results = []
        for item in songlist_json:
            song = Song()
            song.idx = item["idx"]
            song.id = item["id"]
            song.title = item["title_localized"]["en"]
            song.artist = item["artist"]
            song.bpm = item["bpm"]
            song.bpm_base = item["bpm_base"]
            song.set = item["set"]
            song.audio_preview = item["audioPreview"]
            song.audio_preview_end = item["audioPreviewEnd"]
            song.side = item["side"]
            song.version = item["version"]
            song.date = item["date"]
            song.bg = to_db_value(item.get("bg"))
            song.bg_inverse = to_db_value(item.get("bg_inverse"))
            if item.get("bg_daynight"):
                song.bg_day = to_db_value(item["bg_daynight"].get("day"))
                song.bg_night = to_db_value(item["bg_daynight"].get("night"))
            if item.get("source_localized"):
                song.source = item["source_localized"]["en"]
            song.source_copyright = to_db_value(item.get("source_copyright"))
            results.append(song)

            if (
                is_localized(item, "title")
                or is_localized(item, "search_title", append_localized=False)
                or is_localized(item, "search_artist", append_localized=False)
                or is_localized(item, "source")
            ):
                song_localized = SongLocalized(id=song.id)
                set_model_localized_attrs(song_localized, item, "title")
                set_model_localized_attrs(
                    song_localized, item, "search_title", "search_title"
                )
                set_model_localized_attrs(
                    song_localized, item, "search_artist", "search_artist"
                )
                set_model_localized_attrs(song_localized, item, "source")
                results.append(song_localized)

        return results


class SonglistDifficultiesParser(ArcaeaParser):
    def parse(self) -> List[Union[Difficulty, DifficultyLocalized]]:
        songlist_json_root = json.loads(self.read_file_text())

        songlist_json = songlist_json_root["songs"]
        results = []
        for song_item in songlist_json:
            if not song_item.get("difficulties"):
                continue

            for item in song_item["difficulties"]:
                if item["rating"] == 0:
                    continue

                chart = Difficulty(song_id=song_item["id"])
                chart.rating_class = item["ratingClass"]
                chart.rating = item["rating"]
                chart.rating_plus = item.get("ratingPlus") or False
                chart.chart_designer = item["chartDesigner"]
                chart.jacket_desginer = item.get("jacketDesigner") or None
                chart.audio_override = item.get("audioOverride") or False
                chart.jacket_override = item.get("jacketOverride") or False
                chart.jacket_night = item.get("jacketNight") or None
                chart.title = item.get("title_localized", {}).get("en") or None
                chart.artist = item.get("artist") or None
                chart.bg = item.get("bg") or None
                chart.bg_inverse = item.get("bg_inverse")
                chart.bpm = item.get("bpm") or None
                chart.bpm_base = item.get("bpm_base") or None
                chart.version = item.get("version") or None
                chart.date = item.get("date") or None
                results.append(chart)

                if is_localized(item, "title") or is_localized(item, "artist"):
                    chart_localized = DifficultyLocalized(
                        song_id=chart.song_id, rating_class=chart.rating_class
                    )
                    set_model_localized_attrs(chart_localized, item, "title")
                    set_model_localized_attrs(chart_localized, item, "artist")
                    results.append(chart_localized)

        return results



from .config import ConfigBase, Property
from .scores import (
    CalculatedPotential,
    Score,
    ScoreBest,
    ScoreCalculated,
    ScoresBase,
    ScoresViewBase,
)
from .songs import (
    Chart,
    ChartInfo,
    Difficulty,
    DifficultyLocalized,
    Pack,
    PackLocalized,
    Song,
    SongLocalized,
    SongsBase,
    SongsViewBase,
)



# pylint: disable=too-few-public-methods

from sqlalchemy import TEXT
from sqlalchemy.orm import DeclarativeBase, Mapped, mapped_column

from .common import ReprHelper

__all__ = [
    "ConfigBase",
    "Property",
]


class ConfigBase(DeclarativeBase, ReprHelper):
    pass


class Property(ConfigBase):
    __tablename__ = "properties"

    key: Mapped[str] = mapped_column(TEXT(), primary_key=True)
    value: Mapped[str] = mapped_column(TEXT())



import contextlib
import json
import math
import time
from os import PathLike
from typing import Any, List, Optional, Union

from sqlalchemy.orm import DeclarativeBase, Session


def fix_timestamp(timestamp: int) -> Union[int, None]:
    """
    Some of the `date` column in st3 are strangely truncated. For example,
    a `1670283375` may be truncated to `167028`, even `1`. Yes, a single `1`.

    To properly handle this situation, we check the timestamp's digits.
    If `digits < 5`, we treat this timestamp as a `None`. Otherwise, we try to
    fix the timestamp.

    :param timestamp: a POSIX timestamp
    :return: `None` if the timestamp's digits < 5, otherwise a fixed POSIX timestamp
    """
    # find digit length from https://stackoverflow.com/a/2189827/16484891
    # CC BY-SA 2.5
    # this might give incorrect result when timestamp > 999999999999997,
    # see https://stackoverflow.com/a/28883802/16484891 (CC BY-SA 4.0).
    # but that's way too later than 9999-12-31 23:59:59, 253402271999,
    # I don't think Arcaea would still be an active updated game by then.
    # so don't mind those small issues, just use this.
    digits = int(math.log10(abs(timestamp))) + 1 if timestamp != 0 else 1
    if digits < 5:
        return None
    timestamp_str = str(timestamp)
    current_timestamp_digits = int(math.log10(int(time.time()))) + 1
    timestamp_str = timestamp_str.ljust(current_timestamp_digits, "0")
    return int(timestamp_str, 10)


def to_db_value(val: Any) -> Any:
    if not val:
        return None
    return json.dumps(val, ensure_ascii=False) if isinstance(val, list) else val


def is_localized(item: dict, key: str, append_localized: bool = True):
    item_key = f"{key}_localized" if append_localized else key
    subitem: Optional[dict] = item.get(item_key)
    return subitem and (
        subitem.get("ja")
        or subitem.get("ko")
        or subitem.get("zh-Hant")
        or subitem.get("zh-Hans")
    )


def set_model_localized_attrs(
    model: DeclarativeBase, item: dict, model_key: str, item_key: Optional[str] = None
):
    if item_key is None:
        item_key = f"{model_key}_localized"
    subitem: dict = item.get(item_key, {})
    if not subitem:
        return
    setattr(model, f"{model_key}_ja", to_db_value(subitem.get("ja")))
    setattr(model, f"{model_key}_ko", to_db_value(subitem.get("ko")))
    setattr(model, f"{model_key}_zh_hans", to_db_value(subitem.get("zh-Hans")))
    setattr(model, f"{model_key}_zh_hant", to_db_value(subitem.get("zh-Hant")))


class ArcaeaParser:
    def __init__(self, filepath: Union[str, bytes, PathLike]):
        self.filepath = filepath

    def read_file_text(self):
        file_handle = None

        with contextlib.suppress(TypeError):
            # original open
            file_handle = open(self.filepath, "r", encoding="utf-8")

        if file_handle is None:
            try:
                # or maybe a `pathlib.Path` subset
                # or an `importlib.resources.abc.Traversable` like object
                # e.g. `zipfile.Path`
                file_handle = self.filepath.open(mode="r", encoding="utf-8")  # type: ignore
            except Exception as e:
                raise ValueError("Invalid `filepath`.") from e

        with file_handle:
            return file_handle.read()

    def parse(self) -> List[DeclarativeBase]:
        raise NotImplementedError()

    def write_database(self, session: Session):
        results = self.parse()
        for result in results:
            session.merge(result)



from typing import Optional

RATING_CLASS_TEXT_MAP = {
    0: "Past",
    1: "Present",
    2: "Future",
    3: "Beyond",
    4: "Eternal",
}

RATING_CLASS_SHORT_TEXT_MAP = {
    0: "PST",
    1: "PRS",
    2: "FTR",
    3: "BYD",
    4: "ETR",
}


def rating_class_to_text(rating_class: int) -> Optional[str]:
    return RATING_CLASS_TEXT_MAP.get(rating_class)


def rating_class_to_short_text(rating_class: int) -> Optional[str]:
    return RATING_CLASS_SHORT_TEXT_MAP.get(rating_class)



from typing import Optional, Union

from sqlalchemy import select
from sqlalchemy.orm import Session

from ...models import CalculatedPotential, ScoreBest, ScoreCalculated
from .account import AndrealImageGeneratorAccount


class AndrealImageGeneratorApiDataConverter:
    def __init__(
        self,
        session: Session,
        account: AndrealImageGeneratorAccount = AndrealImageGeneratorAccount(),
    ):
        self.session = session
        self.account = account

    def account_info(self):
        return {
            "code": self.account.code,
            "name": self.account.name,
            "is_char_uncapped": self.account.character_uncapped,
            "rating": self.account.rating,
            "character": self.account.character,
        }

    def score(self, score: Union[ScoreCalculated, ScoreBest]):
        return {
            "score": score.score,
            "health": 75,
            "rating": score.potential,
            "song_id": score.song_id,
            "modifier": score.modifier or 0,
            "difficulty": score.rating_class,
            "clear_type": score.clear_type or 1,
            "best_clear_type": score.clear_type or 1,
            "time_played": score.date * 1000 if score.date else 0,
            "near_count": score.far,
            "miss_count": score.lost,
            "perfect_count": score.pure,
            "shiny_perfect_count": score.shiny_pure,
        }

    def user_info(self, score: Optional[ScoreCalculated] = None):
        if not score:
            score = self.session.scalar(
                select(ScoreCalculated).order_by(ScoreCalculated.date.desc()).limit(1)
            )
        if not score:
            raise ValueError("No score available.")

        return {
            "content": {
                "account_info": self.account_info(),
                "recent_score": [self.score(score)],
            }
        }

    def user_best(self, song_id: str, rating_class: int):
        score = self.session.scalar(
            select(ScoreBest).where(
                (ScoreBest.song_id == song_id)
                & (ScoreBest.rating_class == rating_class)
            )
        )
        if not score:
            raise ValueError("No score available.")

        return {
            "content": {
                "account_info": self.account_info(),
                "record": self.score(score),
            }
        }

    def user_best30(self):
        scores = list(
            self.session.scalars(
                select(ScoreBest).order_by(ScoreBest.potential.desc()).limit(40)
            )
        )
        if not scores:
            raise ValueError("No score available.")
        best30_avg = self.session.scalar(select(CalculatedPotential.b30))

        best30_overflow = (
            [self.score(score) for score in scores[30:40]] if len(scores) > 30 else []
        )

        return {
            "content": {
                "account_info": self.account_info(),
                "best30_avg": best30_avg,
                "best30_list": [self.score(score) for score in scores[:30]],
                "best30_overflow": best30_overflow,
            }
        }



from dataclasses import dataclass
from decimal import Decimal
from math import floor
from typing import Tuple, Union


def calculate_score_range(notes: int, pure: int, far: int):
    single_note_score = 10000000 / Decimal(notes)

    actual_score = floor(
        single_note_score * pure + single_note_score * Decimal(0.5) * far
    )
    return (actual_score, actual_score + pure)


def calculate_score_modifier(score: int) -> Decimal:
    if score >= 10000000:
        return Decimal(2)
    if score >= 9800000:
        return Decimal(1) + (Decimal(score - 9800000) / 200000)
    return Decimal(score - 9500000) / 300000


def calculate_play_rating(constant: int, score: int) -> Decimal:
    score_modifier = calculate_score_modifier(score)
    return max(Decimal(0), Decimal(constant) / 10 + score_modifier)


def calculate_shiny_pure(notes: int, score: int, pure: int, far: int) -> int:
    single_note_score = 10000000 / Decimal(notes)
    actual_score = single_note_score * pure + single_note_score * Decimal(0.5) * far
    return score - floor(actual_score)


@dataclass
class ConstantsFromPlayRatingResult:
    # pylint: disable=invalid-name
    EXPlus: Tuple[Decimal, Decimal]
    EX: Tuple[Decimal, Decimal]
    AA: Tuple[Decimal, Decimal]
    A: Tuple[Decimal, Decimal]
    B: Tuple[Decimal, Decimal]
    C: Tuple[Decimal, Decimal]


def calculate_constants_from_play_rating(play_rating: Union[Decimal, str, float, int]):
    # pylint: disable=no-value-for-parameter

    play_rating = Decimal(play_rating)

    ranges = []
    for upper_score, lower_score in [
        (10000000, 9900000),
        (9899999, 9800000),
        (9799999, 9500000),
        (9499999, 9200000),
        (9199999, 8900000),
        (8899999, 8600000),
    ]:
        upper_score_modifier = calculate_score_modifier(upper_score)
        lower_score_modifier = calculate_score_modifier(lower_score)
        ranges.append(
            (play_rating - upper_score_modifier, play_rating - lower_score_modifier)
        )

    return ConstantsFromPlayRatingResult(*ranges)



# pylint: disable=too-few-public-methods

from sqlalchemy.orm import DeclarativeBase
from sqlalchemy.orm.exc import DetachedInstanceError


class ReprHelper:
    # pylint: disable=no-member

    def _repr(self, **kwargs) -> str:
        """
        Helper for __repr__

        https://stackoverflow.com/a/55749579/16484891

        CC BY-SA 4.0
        """
        field_strings = []
        at_least_one_attached_attribute = False
        for key, field in kwargs.items():
            try:
                field_strings.append(f"{key}={field!r}")
            except DetachedInstanceError:
                field_strings.append(f"{key}=DetachedInstanceError")
            else:
                at_least_one_attached_attribute = True
        if at_least_one_attached_attribute:
            return f"<{self.__class__.__name__}({','.join(field_strings)})>"
        return f"<{self.__class__.__name__} {id(self)}>"

    def __repr__(self):
        if isinstance(self, DeclarativeBase):
            return self._repr(
                **{c.key: getattr(self, c.key) for c in self.__table__.columns}
            )
        return super().__repr__()



import logging
import sqlite3
from typing import List

from sqlalchemy import select
from sqlalchemy.orm import Session

from ...models.scores import Score
from .common import ArcaeaParser, fix_timestamp

logger = logging.getLogger(__name__)


class St3ScoreParser(ArcaeaParser):
    def parse(self) -> List[Score]:
        items = []
        with sqlite3.connect(self.filepath) as st3_conn:
            cursor = st3_conn.cursor()
            db_scores = cursor.execute(
                "SELECT songId, songDifficulty, score, perfectCount, nearCount, missCount, "
                "date, modifier FROM scores"
            ).fetchall()
            for (
                song_id,
                rating_class,
                score,
                pure,
                far,
                lost,
                date,
                modifier,
            ) in db_scores:
                clear_type = cursor.execute(
                    "SELECT clearType FROM cleartypes WHERE songId = ? AND songDifficulty = ?",
                    (song_id, rating_class),
                ).fetchone()[0]

                items.append(
                    Score(
                        song_id=song_id,
                        rating_class=rating_class,
                        score=score,
                        pure=pure,
                        far=far,
                        lost=lost,
                        date=fix_timestamp(date),
                        modifier=modifier,
                        clear_type=clear_type,
                        comment="Parsed from st3",
                    )
                )

        return items

    def write_database(self, session: Session, *, skip_duplicate=True):
        parsed_scores = self.parse()
        for parsed_score in parsed_scores:
            query_score = session.scalar(
                select(Score).where(
                    (Score.song_id == parsed_score.song_id)
                    & (Score.rating_class == parsed_score.rating_class)
                    & (Score.score == parsed_score.score)
                )
            )

            if query_score and skip_duplicate:
                logger.info(
                    "%r skipped because potential duplicate item %r found.",
                    parsed_score,
                    query_score,
                )
                continue
            session.add(parsed_score)



class AndrealImageGeneratorAccount:
    def __init__(
        self,
        name: str = "Player",
        code: int = 123456789,
        rating: int = -1,
        character: int = 5,
        character_uncapped: bool = False,
    ):
        self.name = name
        self.code = code
        self.rating = rating
        self.character = character
        self.character_uncapped = character_uncapped



# pylint: disable=too-few-public-methods, duplicate-code

from typing import Optional

from sqlalchemy import TEXT, case, func, inspect, select, text
from sqlalchemy.orm import DeclarativeBase, Mapped, mapped_column
from sqlalchemy_utils import create_view

from .common import ReprHelper
from .songs import ChartInfo, Difficulty

__all__ = [
    "ScoresBase",
    "Score",
    "ScoresViewBase",
    "ScoreCalculated",
    "ScoreBest",
    "CalculatedPotential",
]


class ScoresBase(DeclarativeBase, ReprHelper):
    pass


class Score(ScoresBase):
    __tablename__ = "scores"

    id: Mapped[int] = mapped_column(autoincrement=True, primary_key=True)
    song_id: Mapped[str] = mapped_column(TEXT())
    rating_class: Mapped[int]
    score: Mapped[int]
    pure: Mapped[Optional[int]]
    far: Mapped[Optional[int]]
    lost: Mapped[Optional[int]]
    date: Mapped[Optional[int]]
    max_recall: Mapped[Optional[int]]
    modifier: Mapped[Optional[int]] = mapped_column(
        comment="0: NORMAL, 1: EASY, 2: HARD"
    )
    clear_type: Mapped[Optional[int]] = mapped_column(
        comment="0: TRACK LOST, 1: NORMAL CLEAR, 2: FULL RECALL, "
        "3: PURE MEMORY, 4: EASY CLEAR, 5: HARD CLEAR"
    )
    comment: Mapped[Optional[str]] = mapped_column(TEXT())


# How to create an SQL View with SQLAlchemy?
# https://stackoverflow.com/a/53253105/16484891
# CC BY-SA 4.0


class ScoresViewBase(DeclarativeBase, ReprHelper):
    pass


class ScoreCalculated(ScoresViewBase):
    __tablename__ = "scores_calculated"

    id: Mapped[int]
    song_id: Mapped[str]
    rating_class: Mapped[int]
    score: Mapped[int]
    pure: Mapped[Optional[int]]
    shiny_pure: Mapped[Optional[int]]
    far: Mapped[Optional[int]]
    lost: Mapped[Optional[int]]
    date: Mapped[Optional[int]]
    max_recall: Mapped[Optional[int]]
    modifier: Mapped[Optional[int]]
    clear_type: Mapped[Optional[int]]
    potential: Mapped[float]
    comment: Mapped[Optional[str]]

    __table__ = create_view(
        name=__tablename__,
        selectable=select(
            Score.id,
            Difficulty.song_id,
            Difficulty.rating_class,
            Score.score,
            Score.pure,
            (
                case(
                    (
                        (
                            ChartInfo.notes.is_not(None)
                            & Score.pure.is_not(None)
                            & Score.far.is_not(None)
                            & (ChartInfo.notes != 0)
                        ),
                        Score.score
                        - func.floor(
                            (Score.pure * 10000000.0 / ChartInfo.notes)
                            + (Score.far * 0.5 * 10000000.0 / ChartInfo.notes)
                        ),
                    ),
                    else_=text("NULL"),
                )
            ).label("shiny_pure"),
            Score.far,
            Score.lost,
            Score.date,
            Score.max_recall,
            Score.modifier,
            Score.clear_type,
            case(
                (Score.score >= 10000000, ChartInfo.constant / 10.0 + 2),
                (
                    Score.score >= 9800000,
                    ChartInfo.constant / 10.0 + 1 + (Score.score - 9800000) / 200000.0,
                ),
                else_=func.max(
                    (ChartInfo.constant / 10.0) + (Score.score - 9500000) / 300000.0,
                    0,
                ),
            ).label("potential"),
            Score.comment,
        )
        .select_from(Difficulty)
        .join(
            ChartInfo,
            (Difficulty.song_id == ChartInfo.song_id)
            & (Difficulty.rating_class == ChartInfo.rating_class),
        )
        .join(
            Score,
            (Difficulty.song_id == Score.song_id)
            & (Difficulty.rating_class == Score.rating_class),
        ),
        metadata=ScoresViewBase.metadata,
        cascade_on_drop=False,
    )


class ScoreBest(ScoresViewBase):
    __tablename__ = "scores_best"

    id: Mapped[int]
    song_id: Mapped[str]
    rating_class: Mapped[int]
    score: Mapped[int]
    pure: Mapped[Optional[int]]
    shiny_pure: Mapped[Optional[int]]
    far: Mapped[Optional[int]]
    lost: Mapped[Optional[int]]
    date: Mapped[Optional[int]]
    max_recall: Mapped[Optional[int]]
    modifier: Mapped[Optional[int]]
    clear_type: Mapped[Optional[int]]
    potential: Mapped[float]
    comment: Mapped[Optional[str]]

    __table__ = create_view(
        name=__tablename__,
        selectable=select(
            *[
                col
                for col in inspect(ScoreCalculated).columns
                if col.name != "potential"
            ],
            func.max(ScoreCalculated.potential).label("potential"),
        )
        .select_from(ScoreCalculated)
        .group_by(ScoreCalculated.song_id, ScoreCalculated.rating_class)
        .order_by(ScoreCalculated.potential.desc()),
        metadata=ScoresViewBase.metadata,
        cascade_on_drop=False,
    )


class CalculatedPotential(ScoresViewBase):
    __tablename__ = "calculated_potential"

    b30: Mapped[float]

    _select_bests_subquery = (
        select(ScoreBest.potential.label("b30_sum"))
        .order_by(ScoreBest.potential.desc())
        .limit(30)
        .subquery()
    )
    __table__ = create_view(
        name=__tablename__,
        selectable=select(func.avg(_select_bests_subquery.c.b30_sum).label("b30")),
        metadata=ScoresViewBase.metadata,
        cascade_on_drop=False,
    )



from decimal import Decimal
from typing import Literal, Optional, Union


class PlayResult:
    def __init__(
        self,
        *,
        play_rating: Union[Decimal, str, float, int],
        partner_step: Union[Decimal, str, float, int],
    ):
        self.__play_rating = play_rating
        self.__partner_step = partner_step

    @property
    def play_rating(self):
        return Decimal(self.__play_rating)

    @property
    def partner_step(self):
        return Decimal(self.__partner_step)


class PartnerBonus:
    def __init__(
        self,
        *,
        step_bonus: Union[Decimal, str, float, int] = Decimal("0.0"),
        final_multiplier: Union[Decimal, str, float, int] = Decimal("1.0"),
    ):
        self.__step_bonus = step_bonus
        self.__final_multiplier = final_multiplier

    @property
    def step_bonus(self):
        return Decimal(self.__step_bonus)

    @property
    def final_multiplier(self):
        return Decimal(self.__final_multiplier)


AwakenedIlithPartnerBonus = PartnerBonus(step_bonus="6.0")
AwakenedEtoPartnerBonus = PartnerBonus(step_bonus="7.0")
AwakenedLunaPartnerBonus = PartnerBonus(step_bonus="7.0")


class AwakenedAyuPartnerBonus(PartnerBonus):
    def __init__(self, step_bonus: Union[Decimal, str, float, int]):
        super().__init__(step_bonus=step_bonus)


AmaneBelowExPartnerBonus = PartnerBonus(final_multiplier="0.5")


class MithraTerceraPartnerBonus(PartnerBonus):
    def __init__(self, step_bonus: int):
        super().__init__(step_bonus=step_bonus)


MayaPartnerBonus = PartnerBonus(final_multiplier="2.0")


class StepBooster:
    def final_value(self) -> Decimal:
        raise NotImplementedError()


class LegacyMapStepBooster(StepBooster):
    def __init__(
        self,
        stamina: Literal[2, 4, 6],
        fragments: Literal[100, 250, 500, None],
    ):
        self.stamina = stamina
        self.fragments = fragments

    @property
    def stamina(self):
        return self.__stamina

    @stamina.setter
    def stamina(self, value: Literal[2, 4, 6]):
        if value not in [2, 4, 6]:
            raise ValueError("stamina can only be one of [2, 4, 6]")
        self.__stamina = value

    @property
    def fragments(self):
        return self.__fragments

    @fragments.setter
    def fragments(self, value: Literal[100, 250, 500, None]):
        if value not in [100, 250, 500, None]:
            raise ValueError("fragments can only be one of [100, 250, 500, None]")
        self.__fragments = value

    def final_value(self) -> Decimal:
        stamina_multiplier = Decimal(self.stamina)
        fragments_multiplier = Decimal(1)
        if self.fragments == 100:
            fragments_multiplier = Decimal("1.1")
        elif self.fragments == 250:
            fragments_multiplier = Decimal("1.25")
        elif self.fragments == 500:
            fragments_multiplier = Decimal("1.5")
        return stamina_multiplier * fragments_multiplier


class MemoriesStepBooster(StepBooster):
    def final_value(self) -> Decimal:
        return Decimal("4.0")


def calculate_step_original(
    play_result: PlayResult,
    *,
    partner_bonus: Optional[PartnerBonus] = None,
    step_booster: Optional[StepBooster] = None,
) -> Decimal:
    ptt = play_result.play_rating
    step = play_result.partner_step
    if partner_bonus:
        partner_bonus_step = partner_bonus.step_bonus
        partner_bonus_multiplier = partner_bonus.final_multiplier
    else:
        partner_bonus_step = Decimal("0")
        partner_bonus_multiplier = Decimal("1.0")

    result = (Decimal("2.45") * ptt.sqrt() + Decimal("2.5")) * (step / 50)
    result += partner_bonus_step
    result *= partner_bonus_multiplier
    if step_booster:
        result *= step_booster.final_value()

    return result


def calculate_step(
    play_result: PlayResult,
    *,
    partner_bonus: Optional[PartnerBonus] = None,
    step_booster: Optional[StepBooster] = None,
) -> Decimal:
    result_original = calculate_step_original(
        play_result, partner_bonus=partner_bonus, step_booster=step_booster
    )

    return round(result_original, 1)


def calculate_play_rating_from_step(
    step: Union[Decimal, str, int, float],
    partner_step_value: Union[Decimal, str, int, float],
    *,
    partner_bonus: Optional[PartnerBonus] = None,
    step_booster: Optional[StepBooster] = None,
):
    step = Decimal(step)
    partner_step_value = Decimal(partner_step_value)

    # get original play result
    if partner_bonus and partner_bonus.final_multiplier:
        step /= partner_bonus.final_multiplier
    if step_booster:
        step /= step_booster.final_value()

    if partner_bonus and partner_bonus.step_bonus:
        step -= partner_bonus.step_bonus

    play_rating_sqrt = (Decimal(50) * step - Decimal("2.5") * partner_step_value) / (
        Decimal("2.45") * partner_step_value
    )
    return play_rating_sqrt**2 if play_rating_sqrt >= 0 else -(play_rating_sqrt**2)
